@@ -6,9 +6,10 @@
 #         Vincent Favre-Nicolin, favre@esrf.fr
 
 __all__ = ['default_processing_unit', 'ImshowRGBA', 'ShowCDI', 'SupportUpdate', 'ScaleObj', 'AutoCorrelationSupport',
-           'FreePU', 'FT', 'IFT', 'ER', 'CF', 'HIO', 'RAAR', 'ML', 'SupportUpdate', 'ScaleObj', 'LLK', 'LLKSupport',
-           'DetwinHIO', 'DetwinRAAR', 'SupportExpand', 'ObjConvolve', 'EstimatePSF']
+           'FreePU', 'FT', 'IFT', 'FourierApplyAmplitude', 'ER', 'CF', 'HIO', 'RAAR', 'GPS', 'ML', 'SupportUpdate',
+           'ScaleObj', 'LLK', 'LLKSupport', 'DetwinHIO', 'DetwinRAAR', 'SupportExpand', 'ObjConvolve', 'EstimatePSF']
 
+import timeit
 import types
 import warnings
 import numpy as np
@@ -307,15 +308,21 @@ class ShowCDI(OperatorCDI):
                 v = np.sum(cdi.get_support(), axis=(1, 2))
                 i = int(round(center_of_mass(np.fft.fftshift(v))[0])) - len(cdi.get_obj()) // 2
             obj = cdi.get_obj()[i]
-            iobs = cdi.get_iobs()[0]
+            iobs = cdi.get_iobs()[0].copy()
             icalc = self.get_icalc(cdi, 0)
             if cdi._support is not None:
                 support = cdi.get_support()[i]
         else:
             obj = cdi.get_obj()
-            iobs = cdi.get_iobs()
+            iobs = cdi.get_iobs().copy()
             icalc = self.get_icalc(cdi)
             support = cdi.get_support()
+
+        tmp = np.logical_and(iobs > -1e30, iobs < 0)
+        if tmp.sum() > 0:
+            # change back free pixels to their real intensity
+            iobs[tmp] = -iobs[tmp] - 1
+
         plt.figure(self.fig_num)
         plt.clf()
         plt.subplot(221)
@@ -570,28 +577,41 @@ class ApplyAmplitude(CPUOperatorCDI):
     Apply the magnitude from an observed intensity, keep the phase.
     """
 
-    def __init__(self, calc_llk=False, scale=None):
+    def __init__(self, calc_llk=False, zero_mask=False, scale_in=1, scale_out=1, confidence_interval_factor=0):
         """
 
         :param calc_llk: if true, the log-likelihood will be calculated and stored in the object
-        :param scale: if a floating-point scalar is given, the entire array (including masked values)
-                      is multiplied by it after applying the magnitude constraint.
+        :param zero_mask: if True, masked pixels (iobs<-1e30) are forced to zero, otherwise the calculated
+                          complex amplitude is kept with an optional scale factor.
+        :param scale_in: a scale factor by which the input values should be multiplied, typically because of FFT
+        :param scale_out: a scale factor by which the output values should be multiplied, typically because of FFT
+        :param confidence_interval_factor: a relaxation factor, with the projection of calculated amplitude being done
+                                           towards the limit of the poisson confidence interval. A value of 1
+                                           corresponds to a 50% confidence interval, a value of 0 corresponds to a
+                                           strict observed amplitude projection. [TODO]
         """
         super(ApplyAmplitude, self).__init__()
         self.calc_llk = calc_llk
-        self.scale = scale
+        self.scale_in = np.float32(scale_in)
+        self.scale_out = np.float32(scale_out)
+        self.zero_mask = np.int8(zero_mask)
+        self.confidence_interval_factor = np.float32(confidence_interval_factor)
 
     def op(self, cdi):
+        if self.scale_in != np.float32(1):
+            cdi._obj *= self.scale_in
         if self.calc_llk:
             cdi = LLK() * cdi
         # TODO: partial coherence convolution
         calc = abs(cdi._obj)
         r = np.sqrt(np.fmax(cdi.iobs, 0)) / np.fmax(1e-12, calc)
-        if cdi.mask is not None:
-            r[cdi.mask > 0] = 1
+        if self.zero_mask:
+            r[cdi.iobs < 0] = 0
+        else:
+            r[cdi.iobs < 0] = 1
         cdi._obj *= r
-        if self.scale is not None:
-            cdi._obj *= self.scale
+        if self.scale_out is not None:
+            cdi._obj *= self.scale_out
         return cdi
 
 
@@ -600,8 +620,8 @@ class FourierApplyAmplitude(CPUOperatorCDI):
     Fourier magnitude operator, performing a Fourier transform, the magnitude projection, and a backward FT.
     """
 
-    def __new__(cls, calc_llk=False):
-        return IFT(scale=False) * ApplyAmplitude(calc_llk=calc_llk) * FT(scale=False)
+    def __new__(cls, calc_llk=False, zero_mask=False):
+        return IFT(scale=False) * ApplyAmplitude(calc_llk=calc_llk, zero_mask=zero_mask) * FT(scale=False)
 
 
 class ERProj(CPUOperatorCDI):
@@ -626,8 +646,65 @@ class ER(CPUOperatorCDI):
     Error reduction cycle
     """
 
-    def __new__(cls, positivity=False, calc_llk=False):
-        return ERProj(positivity=positivity) * FourierApplyAmplitude(calc_llk=calc_llk)
+    def __init__(self, positivity=False, calc_llk=False, nb_cycle=1, show_cdi=False, fig_num=-1, zero_mask=False):
+        """
+
+        :param positivity: apply a positivity restraint
+        :param calc_llk: if True, calculate llk while in Fourier space. If a positive integer is given, llk will be
+                         calculated every calc_llk cycle
+        :param nb_cycle: the number of cycles to perform
+        :param show_cdi: if a positive integer number N, the object & probe will be displayed every N cycle.
+                               By default 0 (no plot)
+        :param fig_num: the number of the figure to plot the object intensity, as for ShowCDI()
+        :param zero_mask: if True, masked pixels (iobs<-1e30) are forced to zero, otherwise the calculated
+                          complex amplitude is kept with an optional scale factor.
+        """
+        super(ER, self).__init__()
+        self.positivity = positivity
+        self.calc_llk = calc_llk
+        self.nb_cycle = nb_cycle
+        self.show_cdi = show_cdi
+        self.fig_num = fig_num
+        self.zero_mask = zero_mask
+
+    def __pow__(self, n):
+        """
+
+        :param n: a strictly positive integer
+        :return: a new ER operator with the number of cycles multiplied by n
+        """
+        assert isinstance(n, int) or isinstance(n, np.integer)
+        return ER(positivity=self.positivity, calc_llk=self.calc_llk, nb_cycle=self.nb_cycle * n,
+                  show_cdi=self.show_cdi, fig_num=self.fig_num, zero_mask=self.zero_mask)
+
+    def op(self, cdi: CDI):
+        t0 = timeit.default_timer()
+        ic_dt = 0
+        for ic in range(self.nb_cycle):
+            calc_llk = False
+            if self.calc_llk:
+                if cdi.cycle % self.calc_llk == 0:
+                    calc_llk = True
+            cdi = ERProj(positivity=self.positivity) * FourierApplyAmplitude(calc_llk=calc_llk,
+                                                                             zero_mask=self.zero_mask) * cdi
+
+            if calc_llk:
+                # Average time/cycle over the last N cycles
+                dt = (timeit.default_timer() - t0) / (ic - ic_dt + 1)
+                ic_dt = ic + 1
+                t0 = timeit.default_timer()
+                llk = cdi.get_llk()
+                algo = 'ER'
+                print("%4s #%3d LLK= %8.2f[%8.2f](p) %8.2f[%8.2f](g) %8.2f[%8.2f](e), nb photons=%e, "
+                      "support:nb=%6d (%6.3f%%) average=%10.2f max=%10.2f, dt/cycle=%5.3fs" % (
+                          algo, cdi.cycle, llk[0], llk[3], llk[1], llk[4], llk[2], llk[5], cdi.nb_photons_calc,
+                          cdi.nb_point_support, cdi.nb_point_support / cdi._obj.size * 100,
+                          np.sqrt(cdi.nb_photons_calc / cdi.nb_point_support), cdi._obj_max, dt))
+            if self.show_cdi:
+                if cdi.cycle % self.show_cdi == 0:
+                    cdi = ShowCDI(fig_num=self.fig_num) * cdi
+            cdi.cycle += 1
+        return cdi
 
 
 class CFProj(CPUOperatorCDI):
@@ -652,8 +729,65 @@ class CF(CPUOperatorCDI):
     Charge flipping cycle
     """
 
-    def __new__(cls, positivity=False, calc_llk=False):
-        return CFProj(positivity=positivity) * FourierApplyAmplitude(calc_llk=calc_llk)
+    def __init__(self, positivity=False, calc_llk=False, nb_cycle=1, show_cdi=False, fig_num=-1, zero_mask=False):
+        """
+
+        :param positivity: apply a positivity restraint
+        :param calc_llk: if True, calculate llk while in Fourier space. If a positive integer is given, llk will be
+                         calculated every calc_llk cycle
+        :param nb_cycle: the number of cycles to perform
+        :param show_cdi: if a positive integer number N, the object & probe will be displayed every N cycle.
+                               By default 0 (no plot)
+        :param fig_num: the number of the figure to plot the object intensity, as for ShowCDI()
+        :param zero_mask: if True, masked pixels (iobs<-1e30) are forced to zero, otherwise the calculated
+                          complex amplitude is kept with an optional scale factor.
+        """
+        super(CF, self).__init__()
+        self.positivity = positivity
+        self.calc_llk = calc_llk
+        self.nb_cycle = nb_cycle
+        self.show_cdi = show_cdi
+        self.fig_num = fig_num
+        self.zero_mask = zero_mask
+
+    def __pow__(self, n):
+        """
+
+        :param n: a strictly positive integer
+        :return: a new CF operator with the number of cycles multiplied by n
+        """
+        assert isinstance(n, int) or isinstance(n, np.integer)
+        return CF(positivity=self.positivity, calc_llk=self.calc_llk, nb_cycle=self.nb_cycle * n,
+                  show_cdi=self.show_cdi, fig_num=self.fig_num, zero_mask=self.zero_mask)
+
+    def op(self, cdi: CDI):
+        t0 = timeit.default_timer()
+        ic_dt = 0
+        for ic in range(self.nb_cycle):
+            calc_llk = False
+            if self.calc_llk:
+                if cdi.cycle % self.calc_llk == 0:
+                    calc_llk = True
+            cdi = CFProj(positivity=self.positivity) * FourierApplyAmplitude(calc_llk=calc_llk,
+                                                                             zero_mask=self.zero_mask) * cdi
+
+            if calc_llk:
+                # Average time/cycle over the last N cycles
+                dt = (timeit.default_timer() - t0) / (ic - ic_dt + 1)
+                ic_dt = ic + 1
+                t0 = timeit.default_timer()
+                llk = cdi.get_llk()
+                algo = 'CF'
+                print("%4s #%3d LLK= %8.2f[%8.2f](p) %8.2f[%8.2f](g) %8.2f[%8.2f](e), nb photons=%e, "
+                      "support:nb=%6d (%6.3f%%) average=%10.2f max=%10.2f, dt/cycle=%5.3fs" % (
+                          algo, cdi.cycle, llk[0], llk[3], llk[1], llk[4], llk[2], llk[5], cdi.nb_photons_calc,
+                          cdi.nb_point_support, cdi.nb_point_support / cdi._obj.size * 100,
+                          np.sqrt(cdi.nb_photons_calc / cdi.nb_point_support), cdi._obj_max, dt))
+            if self.show_cdi:
+                if cdi.cycle % self.show_cdi == 0:
+                    cdi = ShowCDI(fig_num=self.fig_num) * cdi
+            cdi.cycle += 1
+        return cdi
 
 
 class HIOProj(CPUOperatorCDI):
@@ -680,8 +814,67 @@ class HIO(CPUOperatorCDI):
     Hybrid Input-Output reduction cycle
     """
 
-    def __new__(cls, beta=0.9, positivity=False, calc_llk=False):
-        return HIOProj(beta, positivity=positivity) * FourierApplyAmplitude(calc_llk=calc_llk) * CopyToPrevious()
+    def __init__(self, beta=0.9, positivity=False, calc_llk=False, nb_cycle=1, show_cdi=False, fig_num=-1,
+                 zero_mask=False):
+        """
+
+        :param positivity: apply a positivity restraint
+        :param calc_llk: if True, calculate llk while in Fourier space. If a positive integer is given, llk will be
+                         calculated every calc_llk cycle
+        :param nb_cycle: the number of cycles to perform
+        :param show_cdi: if a positive integer number N, the object & probe will be displayed every N cycle.
+                               By default 0 (no plot)
+        :param fig_num: the number of the figure to plot the object intensity, as for ShowCDI()
+        :param zero_mask: if True, masked pixels (iobs<-1e30) are forced to zero, otherwise the calculated
+                          complex amplitude is kept with an optional scale factor.
+        """
+        super(HIO, self).__init__()
+        self.beta = beta
+        self.positivity = positivity
+        self.calc_llk = calc_llk
+        self.nb_cycle = nb_cycle
+        self.show_cdi = show_cdi
+        self.fig_num = fig_num
+        self.zero_mask = zero_mask
+
+    def __pow__(self, n):
+        """
+
+        :param n: a strictly positive integer
+        :return: a new HIO operator with the number of cycles multiplied by n
+        """
+        assert isinstance(n, int) or isinstance(n, np.integer)
+        return HIO(beta=self.beta, positivity=self.positivity, calc_llk=self.calc_llk, nb_cycle=self.nb_cycle * n,
+                   show_cdi=self.show_cdi, fig_num=self.fig_num, zero_mask=self.zero_mask)
+
+    def op(self, cdi: CDI):
+        t0 = timeit.default_timer()
+        ic_dt = 0
+        for ic in range(self.nb_cycle):
+            calc_llk = False
+            if self.calc_llk:
+                if cdi.cycle % self.calc_llk == 0:
+                    calc_llk = True
+            cdi = HIOProj(beta=self.beta, positivity=self.positivity) * FourierApplyAmplitude(
+                calc_llk=calc_llk, zero_mask=self.zero_mask) * CopyToPrevious() * cdi
+
+            if calc_llk:
+                # Average time/cycle over the last N cycles
+                dt = (timeit.default_timer() - t0) / (ic - ic_dt + 1)
+                ic_dt = ic + 1
+                t0 = timeit.default_timer()
+                llk = cdi.get_llk()
+                algo = 'HIO'
+                print("%4s #%3d LLK= %8.2f[%8.2f](p) %8.2f[%8.2f](g) %8.2f[%8.2f](e), nb photons=%e, "
+                      "support:nb=%6d (%6.3f%%) average=%10.2f max=%10.2f, dt/cycle=%5.3fs" % (
+                          algo, cdi.cycle, llk[0], llk[3], llk[1], llk[4], llk[2], llk[5], cdi.nb_photons_calc,
+                          cdi.nb_point_support, cdi.nb_point_support / cdi._obj.size * 100,
+                          np.sqrt(cdi.nb_photons_calc / cdi.nb_point_support), cdi._obj_max, dt))
+            if self.show_cdi:
+                if cdi.cycle % self.show_cdi == 0:
+                    cdi = ShowCDI(fig_num=self.fig_num) * cdi
+            cdi.cycle += 1
+        return cdi
 
 
 class RAARProj(CPUOperatorCDI):
@@ -708,8 +901,207 @@ class RAAR(CPUOperatorCDI):
     RAAR cycle
     """
 
-    def __new__(cls, beta=0.9, positivity=False, calc_llk=False):
-        return RAARProj(beta, positivity=positivity) * FourierApplyAmplitude(calc_llk=calc_llk) * CopyToPrevious()
+    def __init__(self, beta=0.9, positivity=False, calc_llk=False, nb_cycle=1, show_cdi=False, fig_num=-1,
+                 zero_mask=False):
+        """
+
+        :param positivity: apply a positivity restraint
+        :param calc_llk: if True, calculate llk while in Fourier space. If a positive integer is given, llk will be
+                         calculated every calc_llk cycle
+        :param nb_cycle: the number of cycles to perform
+        :param show_cdi: if a positive integer number N, the object & probe will be displayed every N cycle.
+                               By default 0 (no plot)
+        :param fig_num: the number of the figure to plot the object intensity, as for ShowCDI()
+        :param zero_mask: if True, masked pixels (iobs<-1e30) are forced to zero, otherwise the calculated
+                          complex amplitude is kept with an optional scale factor.
+        """
+        super(RAAR, self).__init__()
+        self.beta = beta
+        self.positivity = positivity
+        self.calc_llk = calc_llk
+        self.nb_cycle = nb_cycle
+        self.show_cdi = show_cdi
+        self.fig_num = fig_num
+        self.zero_mask = zero_mask
+
+    def __pow__(self, n):
+        """
+
+        :param n: a strictly positive integer
+        :return: a new HIO operator with the number of cycles multiplied by n
+        """
+        assert isinstance(n, int) or isinstance(n, np.integer)
+        return RAAR(beta=self.beta, positivity=self.positivity, calc_llk=self.calc_llk, nb_cycle=self.nb_cycle * n,
+                    show_cdi=self.show_cdi, fig_num=self.fig_num, zero_mask=self.zero_mask)
+
+    def op(self, cdi: CDI):
+        t0 = timeit.default_timer()
+        ic_dt = 0
+        for ic in range(self.nb_cycle):
+            calc_llk = False
+            if self.calc_llk:
+                if cdi.cycle % self.calc_llk == 0:
+                    calc_llk = True
+
+            cdi = RAARProj(self.beta, positivity=self.positivity) * FourierApplyAmplitude(
+                calc_llk=calc_llk, zero_mask=self.zero_mask) * CopyToPrevious() * cdi
+
+            if calc_llk:
+                # Average time/cycle over the last N cycles
+                dt = (timeit.default_timer() - t0) / (ic - ic_dt + 1)
+                ic_dt = ic + 1
+                t0 = timeit.default_timer()
+                llk = cdi.get_llk()
+                algo = 'RAAR'
+                print("%4s #%3d LLK= %8.2f[%8.2f](p) %8.2f[%8.2f](g) %8.2f[%8.2f](e), nb photons=%e, "
+                      "support:nb=%6d (%6.3f%%) average=%10.2f max=%10.2f, dt/cycle=%5.3fs" % (
+                          algo, cdi.cycle, llk[0], llk[3], llk[1], llk[4], llk[2], llk[5], cdi.nb_photons_calc,
+                          cdi.nb_point_support, cdi.nb_point_support / cdi._obj.size * 100,
+                          np.sqrt(cdi.nb_photons_calc / cdi.nb_point_support), cdi._obj_max, dt))
+            if self.show_cdi:
+                if cdi.cycle % self.show_cdi == 0:
+                    cdi = ShowCDI(fig_num=self.fig_num) * cdi
+            cdi.cycle += 1
+        return cdi
+
+
+class GPS(CPUOperatorCDI):
+    """
+    GPS cycle, according to Pham et al [2019]
+    """
+
+    def __init__(self, inertia=0.05, t=1.0, s=0.9, sigma_f=0, sigma_o=0, positivity=False,
+                 calc_llk=False, nb_cycle=1, show_cdi=False, fig_num=-1, zero_mask=False):
+        """
+        :param inertia: inertia parameter (sigma in original Pham2019 article)
+        :param t: t parameter
+        :param s: s parameter
+        :param sigma_f: Fourier-space smoothing kernel width, in Fourier-space pixel units
+        :param sigma_o: object-space smoothing kernel width, in object-space pixel units
+        :param positivity: apply a positivity restraint
+        :param calc_llk: if True, calculate llk while in Fourier space. If a positive integer is given, llk will be
+                         calculated every calc_llk cycle
+        :param nb_cycle: the number of cycles to perform
+        :param show_cdi: if a positive integer number N, the object & probe will be displayed every N cycle.
+                               By default 0 (no plot)
+        :param fig_num: the number of the figure to plot the object intensity, as for ShowCDI()
+        :param zero_mask: if True, masked pixels (iobs<-1e30) are forced to zero, otherwise the calculated
+                          complex amplitude is kept with an optional scale factor.
+        """
+        super(GPS, self).__init__()
+        self.inertia = np.float32(inertia)
+        self.t = np.float32(t)
+        self.s = np.float32(s)
+        self.sigma_f = np.float32(sigma_f)
+        self.sigma_o = np.float32(sigma_o)
+        self.positivity = positivity
+        self.calc_llk = calc_llk
+        self.nb_cycle = nb_cycle
+        self.show_cdi = show_cdi
+        self.fig_num = fig_num
+        self.zero_mask = zero_mask
+
+    def __pow__(self, n):
+        """
+
+        :param n: a strictly positive integer
+        :return: a new GPS operator with the number of cycles multiplied by n
+        """
+        assert isinstance(n, int) or isinstance(n, np.integer)
+        return GPS(inertia=self.inertia, t=self.t, s=self.s, sigma_f=self.sigma_f, sigma_o=self.sigma_o,
+                   positivity=self.positivity, calc_llk=self.calc_llk, nb_cycle=self.nb_cycle * n,
+                   show_cdi=self.show_cdi, fig_num=self.fig_num, zero_mask=self.zero_mask)
+
+    def op(self, cdi: CDI):
+        t0 = timeit.default_timer()
+        ic_dt = 0
+
+        epsilon = np.float32(self.inertia / (self.inertia + self.t))
+
+        ny, nx = np.int32(cdi._obj.shape[-2]), np.int32(cdi._obj.shape[-1])
+        if cdi._obj.ndim == 3:
+            nz = np.int32(cdi._obj.shape[0])
+        else:
+            nz = np.int32(1)
+
+        # We start in Fourier space (obj = z_0)
+        cdi = FT(scale=True) * cdi
+
+        # z_0 = FT(obj)
+        z = cdi._obj.copy()
+
+        # Start with obj = y_0 = 0
+        cdi._obj.fill(np.complex64(0))
+
+        # Gaussian smoothing arrays in object and Fourier space
+        if cdi.iobs.ndim == 2:
+            ny, nx = cdi.iobs.shape
+            qx = np.fft.fftfreq(nx).astype(np.float32)
+            qy = np.fft.fftfreq(ny).astype(np.float32)
+            qy, qx = np.meshgrid(qy, qx, indexing='ij')
+            qz = 0
+        else:
+            nz, ny, nx = cdi.iobs.shape
+            qx = np.fft.fftfreq(nx).astype(np.float32)
+            qy = np.fft.fftfreq(ny).astype(np.float32)
+            qz = np.fft.fftfreq(nz).astype(np.float32)
+            qz, qy, qx = np.meshgrid(qz, qy, qx, indexing='ij')
+
+        g_o = np.exp(-2 * np.pi ** 2 * self.sigma_o ** 2 * (qx ** 2 + qy ** 2 + qz ** 2))
+        g_f = np.exp(-2 * np.pi ** 2 * self.sigma_f ** 2 * (qx ** 2 + qy ** 2 + qz ** 2))
+
+        for ic in range(self.nb_cycle):
+            calc_llk = False
+            if self.calc_llk:
+                if cdi.cycle % self.calc_llk == 0:
+                    calc_llk = True
+
+            # keep y copy
+            y = cdi._obj.copy()
+
+            cdi = FT(scale=True) * cdi
+
+            # ^z = z_k - t F(y_k)
+            cdi._obj = z - self.t * g_o * cdi._obj
+
+            cdi = ApplyAmplitude(calc_llk=calc_llk, zero_mask=self.zero_mask) * cdi
+
+            # obj = z_k+1 = (1 - epsilon) * sqrt(iobs) * exp(i * arg(^z)) + epsilon * z_k
+            cdi._obj = (1-epsilon) * cdi._obj + epsilon * z
+
+            if calc_llk:
+                # Average time/cycle over the last N cycles
+                dt = (timeit.default_timer() - t0) / (ic - ic_dt + 1)
+                ic_dt = ic + 1
+                t0 = timeit.default_timer()
+                llk = cdi.get_llk()
+                cdi.update_history(mode='llk', dt=dt, algorithm='GPS', verbose=True)
+            else:
+                cdi.history.insert(cdi.cycle, algorithm='GPS')
+            if self.show_cdi:
+                if cdi.cycle % self.show_cdi == 0:
+                    cdi = IFT(scale=True) * cdi
+                    cdi = ShowCDI(fig_num=self.fig_num) * cdi
+                    cdi = FT(scale=True) * cdi
+            cdi.cycle += 1
+
+            if ic < self.nb_cycle - 1:
+                # obj = 2 * z_k+1 - z_k  & store z_k+1 in z
+                cdi._obj, z = 2 * cdi._obj - z, cdi._obj
+
+                cdi = IFT(scale=True) * cdi
+
+                # obj = ^y = proj_support[y_k + s * obj] * G_sigma_f
+                cdi._obj = g_f * (y + self.s * cdi._obj)
+                if self.positivity:
+                    cdi._obj *= g_f * np.logical_or(cdi._support == 0, cdi._obj.real < 0)
+                else:
+                    cdi._obj *= g_f * (cdi._support == 0)
+
+        # Back to object space
+        cdi = IFT(scale=True) * cdi
+
+        return cdi
 
 
 class ML(CPUOperatorCDI):
@@ -717,11 +1109,23 @@ class ML(CPUOperatorCDI):
     Maximum likelihood conjugate gradient minimization
     """
 
-    def __init__(self, reg_fac=1e-2, nb_cycle=1, calc_llk=False):
+    def __init__(self, reg_fac=1e-2, nb_cycle=1, calc_llk=False, show_cdi=False, fig_num=-1):
+        """
+
+        :param reg_fac:
+        :param nb_cycle: the number of cycles to perform
+        :param calc_llk: if True, calculate llk while in Fourier space. If a positive integer is given, llk will be
+                         calculated every calc_llk cycle
+        :param show_cdi: if a positive integer number N, the object & probe will be displayed every N cycle.
+                               By default 0 (no plot)
+        :param fig_num: the number of the figure to plot the object intensity, as for ShowCDI()
+        """
         super(ML, self).__init__()
         self.reg_fac = reg_fac
         self.nb_cycle = nb_cycle
         self.calc_llk = calc_llk
+        self.show_cdi = show_cdi
+        self.fig_num = fig_num
 
     def __pow__(self, n):
         """
@@ -730,7 +1134,8 @@ class ML(CPUOperatorCDI):
         :return: a new ML operator with the number of cycles multiplied by n
         """
         assert isinstance(n, int) or isinstance(n, np.integer)
-        return ML(reg_fac=self.reg_fac, nb_cycle=self.nb_cycle * n, calc_llk=self.calc_llk)
+        return ML(reg_fac=self.reg_fac, nb_cycle=self.nb_cycle * n, calc_llk=self.calc_llk, show_cdi=self.show_cdi,
+                  fig_num=self.fig_num)
 
     def op(self, cdi: CDI):
         obj_grad = np.empty_like(cdi._obj)
@@ -740,7 +1145,7 @@ class ML(CPUOperatorCDI):
         # Use support for regularization
         N = cdi._obj.size
         # Total number of photons
-        Nph = cdi.iobs.sum()
+        Nph = cdi.iobs_sum
         cdi.llk_support_reg_fac = np.float32(self.reg_fac / (8 * N / Nph))
 
         # if self.reg_fac_support>0:
@@ -753,7 +1158,7 @@ class ML(CPUOperatorCDI):
             psi = np.fft.fftn(cdi._obj, norm="ortho")
             icalc = abs(psi) ** 2
             iobs = cdi.iobs.copy()
-            iobs[cdi.mask > 0] = icalc[cdi.mask > 0]
+            iobs[cdi.iobs < 0] = icalc[cdi.iobs < 0]
 
             if self.calc_llk and cycle == (self.nb_cycle - 1):
                 cdi = LLK() * cdi
@@ -797,58 +1202,98 @@ class SupportUpdate(OperatorCDI):
     Update the support
     """
 
-    def __init__(self, threshold_relative=0.2, threshold_percentile=None, smooth_width=3, force_shrink=False):
+    def __init__(self, threshold_relative=0.2, smooth_width=3, force_shrink=False, method='average',
+                 post_expand=None, verbose=False):
         """
         Update support.
 
         Args:
-            threshold_relative: must be between 0 and 1. Only points with object amplitude above a value equal to 
+            threshold_relative: must be between 0 and 1. Only points with object amplitude above a value equal to
                 relative_threshold * reference_value are kept in the support.
                 reference_value can either:
-                - use the fact that when converged, the square norm of the object is equal to the number of 
+                - use the fact that when converged, the square norm of the object is equal to the number of
                 recorded photons (normalized Fourier Transform). Then:
                   reference_value = sqrt((abs(obj)**2).sum()/nb_points_support)
                 - or use threshold_percentile (see below, very slow, deprecated)
-            threshold_percentile: DEPRECATED (too much time-consuming)
-                value between 0 and 100 using np.percentile to calculate the reference_value, 
-                only taking into account the number of points in the support.
             smooth_width: smooth the object amplitude using a gaussian of this width before calculating new support
+                          If this is a scalar, the smooth width is fixed to this value.
+                          If this is a 3-value tuple (or list or array), i.e. 'smooth_width=2,0.5,600', the smooth width
+                          will vary with the number of cycles recorded in the CDI object (as cdi.cycle), varying
+                          exponentially from the first to the second value over the number of cycles specified by the
+                          last value.
+                          With 'smooth_width=a,b,nb':
+                               smooth_width = a * exp(-cdi.cycle/nb*log(b/a)) if cdi.cycle < nb
+                               smooth_width = b if cdi.cycle >= nb
             force_shrink: if True, the support can only shrink
+            method: either 'max' or 'average' (default), the threshold will be relative to either the maximum amplitude
+                    in the object, or the average amplitude (calculated as total intensity, divided by support size)
+            post_expand=1: after the new support has been calculated, it can be processed using the SupportExpand
+                           operator, either one or multiple times, in order to 'clean' the support:
+                           - 'post_expand=1' will expand the support by 1 pixel
+                           - 'post_expand=-1' will shrink the support by 1 pixel
+                           - 'post_expand=(-1,1)' will shrink and then expand the support by 1 pixel
+                           - 'post_expand=(-2,3)' will shrink and then expand the support by respectively 2 and 3 pixels
+            verbose: if True, print number of points in support
         Returns:
             Nothing. self._support is updated
         """
         super(SupportUpdate, self).__init__()
-        self.threshold_relative = threshold_relative
-        self.threshold_percentile = threshold_percentile
         self.smooth_width = smooth_width
-        self.force_shrink = force_shrink
+        self.threshold_relative = threshold_relative
+        self.force_shrink = np.bool(force_shrink)
+        self.method = method
+        self.verbose = verbose
+        if isinstance(post_expand, int) or isinstance(post_expand, np.integer):
+            self.post_expand = (post_expand,)
+        else:
+            self.post_expand = post_expand
 
     def op(self, cdi):
-        # TODO: faster, avoid recomputing iobs.sum, and use a GPU kernel for convolution/thresholding.
-        tmpobj = np.fft.fftshift(np.abs(cdi.get_obj()).astype(np.float32))
-        tmpobj = np.fft.fftshift(gaussian_filter(tmpobj, self.smooth_width))
-        support = cdi.get_support()
-        if self.threshold_percentile is not None:
-            # DEPRECATED
-            # Threshold = percentile of object amplitude, cutting at half the value allowing 80% of points
-            # in the support
-            t = support.sum() / float(support.size)
-            percentile = (1 - t) * 100 + self.threshold_percentile * t
-            if percentile < 1:
-                print("support_update: percentile < 1 !! (%f), threshold_relative=%f, threshold_percentile=%f" %
-                      (percentile, self.threshold_relative, self.threshold_percentile))
-                percentile = 1
-            if percentile > 100:
-                percentile = 100
-                print("support_update: percentile > 100 !! (%f), threshold_relative=%f, threshold_percentile=%f" %
-                      (percentile, self.threshold_relative, self.threshold_percentile))
-            thr = np.percentile(tmpobj, percentile) * self.threshold_relative
+        if np.isscalar(self.smooth_width):
+            smooth_width = self.smooth_width
         else:
-            thr = self.threshold_relative * np.sqrt((abs(cdi.get_obj()) ** 2).sum() / support.sum())
+            a, b, nb = self.smooth_width
+            i = cdi.cycle
+            if i < nb:
+                smooth_width = a * np.exp(-i / nb * np.log(a / b))
+            else:
+                smooth_width = b
+        tmpobj = np.fft.fftshift(np.abs(cdi.get_obj()).astype(np.float32))
+        tmpobj = np.fft.fftshift(gaussian_filter(tmpobj, smooth_width))
+        support = cdi.get_support()
+
+        # Get total number of photons and maximum intensity in the support
+        tmp = (np.abs(support * cdi.get_obj()) ** 2)
+        max_icalc_support = tmp.max()
+        nb_ph_support = tmp.sum()
+        cdi._obj_max = np.sqrt(max_icalc_support)
+
+        # Get average amplitude and maximum intensity in the support, from the convolved amplitude
+        tmp = (np.abs(support * tmpobj))
+        max_abs_support = tmp.max()
+        av_abs_support = tmp.sum() / cdi.nb_point_support
+
+        # Threshold (from average)
+        if self.method == 'max':
+            thr = self.threshold_relative * np.float32(max_abs_support)
+        else:
+            thr = self.threshold_relative * np.float32(av_abs_support)
+        # Update support and compute the new number of points in the support
         if self.force_shrink:
             support *= tmpobj > thr
         else:
             support = (tmpobj > thr).astype(np.int8)
+
+        if self.post_expand is not None:
+            for n in self.post_expand:
+                cdi = SupportExpand(n=n, update_nb_points_support=False) * cdi
+
+        nb = support.sum()
+        if self.verbose:
+            print("Nb points in support: %d (%6.3f%%), threshold=%8f  (%6.3f), nb photons=%10e"
+                  % (nb, nb / cdi._obj.size * 100, thr, self.threshold_relative, nb_ph_support))
+        cdi.nb_point_support = nb
+
         cdi.set_support(support)
         return cdi
 
@@ -858,36 +1303,43 @@ class ScaleObj(OperatorCDI):
     Scale the object according to the observed intensity. The scaling is either made against the amplitudes,
     the intensities, or the weighted intensities.
     This is only useful if a mask is used - the scale factor effectively only applies to masked intensities.
-    :param method: 'I' (intensities), 'F' (amplitudes), 'wI' (weighted intensities)
+    :param method: 'I' (intensities), 'F' (amplitudes), 'wI' (weighted intensities), 'P' Poisson
     :return: nothing. The object is scaled to best match the intensities.
     """
 
     # TODO: move this to GPU. Integrate with FourierApplyAmplitude operator.
-    def __init__(self, method='I'):
+    def __init__(self, method='I', verbose=False):
         """
-        :param method: 'I' (intensities), 'F' (amplitudes), 'wI' (weighted intensities)
+        :param method: 'I' (intensities), 'F' (amplitudes), 'wI' (weighted intensities), 'P' (Poisson)
+        :param verbose: if True, print the scale factor
         """
         super(ScaleObj, self).__init__()
         self.method = method
+        self.verbose = verbose
 
     def op(self, cdi):
         if self.method.lower() == 'f':
             # Scale the object to match Fourier amplitudes
-            tmpcalc = np.abs(np.fft.fftn(cdi._obj, norm=None)) * (cdi.mask == 0)
+            tmpcalc = np.abs(np.fft.fftn(cdi._obj, norm=None)) * (cdi.iobs >= 0)
             tmpobs = np.sqrt(np.abs(cdi.iobs))
             scale = (tmpcalc * tmpobs).sum() / (tmpcalc ** 2).sum()
         elif self.method.lower() == 'i':
             # Scale object to match Fourier intensities
-            tmpcalc = np.abs(np.fft.fftn(cdi._obj, norm=None)) ** 2 * (cdi.mask == 0)
+            tmpcalc = np.abs(np.fft.fftn(cdi._obj, norm=None)) ** 2 * (cdi.iobs >= 0)
+            scale = np.sqrt((tmpcalc * cdi.iobs).sum() / (tmpcalc ** 2).sum())
+        elif self.method.lower() == 'p':
+            # Scale object to match Poisson statistics
+            tmpcalc = np.abs(np.fft.fftn(cdi._obj, norm=None)) ** 2 * (cdi.iobs >= 0)
             scale = np.sqrt((tmpcalc * cdi.iobs).sum() / (tmpcalc ** 2).sum())
         else:
             # Scale object to match weighted intensities
             # Weight: 1 for null intensities, zero for masked pixels
-            w = (1 / (np.abs(cdi.iobs) + 1e-6) * (cdi.iobs > 1e-6) + (cdi.iobs <= 1e-6)) * (cdi.mask == 0)
+            w = (1 / (np.abs(cdi.iobs) + 1e-6) * (cdi.iobs > 1e-6) + (cdi.iobs <= 1e-6)) * (cdi.iobs >= 0)
             tmpcalc = np.abs(np.fft.fftn(cdi._obj, norm=None)) ** 2
             scale = np.sqrt((w * tmpcalc * cdi.iobs).sum() / (w * tmpcalc ** 2).sum())
         cdi._obj *= scale
-        print("Scaled object by: %f" % (scale))
+        if self.verbose:
+            print("Scaled object by: %f" % (scale))
         return cdi
 
 
@@ -900,6 +1352,15 @@ class LLK(CPUOperatorCDI):
     total calculated intensity (including in masked pixels).
     """
 
+    def __init__(self, scale=1.0):
+        """
+
+        :param scale: the scale factor to be applied to the calculated amplitude before evaluating the
+                      log-likelihood. The calculated amplitudes are left unmodified.
+        """
+        super(LLK, self).__init__()
+        self.scale = np.float32(scale ** 2)
+
     def op(self, cdi):
         need_ft = cdi.in_object_space()
 
@@ -907,9 +1368,9 @@ class LLK(CPUOperatorCDI):
             cdi = FT() * cdi
 
         # TODO: add support for convolution with partial coherence kernel
-        calc = abs(cdi._obj) ** 2
+        calc = np.abs(self.scale * cdi._obj) ** 2
         cdi.nb_photons_calc = calc.sum()
-        calc[cdi.mask > 0] = cdi.iobs[cdi.mask > 0]
+        calc[cdi.iobs < 0] = cdi.iobs[cdi.iobs < 0]
         cdi.llk_poisson = llk_poisson(cdi.iobs, calc).sum()
         cdi.llk_gaussian = llk_gaussian(cdi.iobs, calc).sum()
         cdi.llk_euclidian = llk_euclidian(cdi.iobs, calc).sum()
@@ -1074,8 +1535,7 @@ class EstimatePSF(CPUOperatorCDI):
         n = cdi._k_psf_n  # Kernel size = 2*n+1
 
         icalc = abs(cdi.get_obj(shift=True)).astype(np.float32) ** 2
-        mask = np.fft.fftshift(cdi.mask)
-        iobs = (cdi.get_iobs(shift=True) * (mask == 0) + icalc * (mask > 0)).astype(np.float32)
+        iobs = (cdi.get_iobs(shift=True) * (cdi.iobs < 0) + icalc * (cdi.iobs < 0)).astype(np.float32)
 
         # Use only the center of the data to estimate the PSF - much faster !
         if cdi._obj.ndim == 2:
