@@ -187,7 +187,7 @@ class ProcessingUnit(object):
                       DeprecationWarning)
         return self.select_gpu(gpu_name=gpu_name, gpu_rank=gpu_rank, ranking=ranking, language=language)
 
-    def select_gpu(self, gpu_name=None, gpu_rank=0, ranking="fft_speed", language=None, verbose=True):
+    def select_gpu(self, gpu_name=None, gpu_rank=0, ranking="fft_speed", language=None, verbose=False):
 
         """
         Automatically select the fastest GPU, either from CUDA or OpenCL.
@@ -200,7 +200,8 @@ class ProcessingUnit(object):
 
         This will call set_device() with the selected device.
 
-        :param gpu_name: a (sub)string matching the name of the gpu to be used
+        :param gpu_name: a (sub)string matching the name of the gpu to be used. This can also be a list of
+                         acceptable gpu names.
         :param gpu_rank: the rank of the GPU to use. If ranking is 'fft_speed', this will select from the fastest
                          GPU. If ranking is 'order', it will just find the GPU matching the name, in the order
                          they are found.
@@ -208,7 +209,7 @@ class ProcessingUnit(object):
                         ranking. If ranking is 'order', they will be listed as they are found by CUDA/OpenCL.
                         This is useful to select a GPU just based on its name, without the overhead of the FFT test
         :param language: either 'opencl', 'cuda', 'cpu', or None. If None, the preferred language is cuda>opencl>cpu
-        :param verbose: True (default) or False for verbosity
+        :param verbose: True or False (default) for verbosity
         :return: True if a GPU (or CPU) could be selected. An exception is returned if no GPU was found,
                  unless language was 'cpu'.
         :raise Exception: if no GPU could be selected, and the language was not 'cpu'. The message will indicate
@@ -258,7 +259,7 @@ class ProcessingUnit(object):
                     if len(benchmark_results) == 0:
                         s = 'Desired GPU language is CUDA, but no device found !'
                         if gpu_name is not None:
-                            s += ' [gpu_name=%s]' % gpu_name
+                            s += ' [gpu_name=%s]' % str(gpu_name)
                         raise Exception(s)
 
         test_opencl = False
@@ -280,7 +281,7 @@ class ProcessingUnit(object):
                     if len(benchmark_results) == 0:
                         s = 'Desired GPU language is OpenCL, but no device found !'
                         if gpu_name is not None:
-                            s += ' [gpu_name=%s]' % gpu_name
+                            s += ' [gpu_name=%s]' % str(gpu_name)
                         raise Exception(s)
 
         if language is not None:
@@ -288,7 +289,7 @@ class ProcessingUnit(object):
                 return self.set_device(test_fft=False, verbose=verbose)
 
         if gpu_name is not None:
-            if 'cpu' in gpu_name.lower():
+            if 'cpu' in str(gpu_name).lower():
                 return self.set_device(test_fft=False, verbose=verbose)
 
         if len(benchmark_results):
@@ -296,7 +297,7 @@ class ProcessingUnit(object):
                 benchmark_results = sorted(benchmark_results, key=lambda t: -t[1])
             if len(benchmark_results) <= gpu_rank:
                 if gpu_name is not None:
-                    s = ' [gpu_name=%s]' % gpu_name
+                    s = ' [gpu_name=%s]' % str(gpu_name)
                 else:
                     s = ''
                 raise Exception('Desired GPU%s rank is %d, but only %d GPU found !' % (s, len(benchmark_results),
@@ -306,7 +307,7 @@ class ProcessingUnit(object):
         s = "Could not find a suitable GPU. Please check GPU name or CUDA/OpenCL installation"
 
         if gpu_name is not None:
-            s += " [name=%s]" % gpu_name
+            s += " [name=%s]" % str(gpu_name)
         raise Exception(s)
 
     def finish(self):
