@@ -7,41 +7,110 @@
 
 __all__ = ['Bragg2DPtychoData', 'Bragg2DPtycho', 'OperatorBragg2DPtycho']
 
-
 import numpy as np
 from scipy.interpolate import RegularGridInterpolator
-from ...operator import Operator
+from ...operator import Operator, has_attr_not_none
 
 
 # TODO: Merge as much as possible with the 3D Bragg case (almost identical)
 
+def rotation_matrix(axis, angle):
+    """
+    Creates a numpy rotation matrix. The convention is the NeXus one so that a positive rotation of +pi/2:
+    - with axis='x', transforms +y into z
+    - with axis='y', transforms +z into x
+    - with axis='z', transforms +x into y
+    :param axis: the rotation axis, either 'x', 'y' or 'z'
+    :param angle: the rotation angle in radians
+    :return: the rotation matrix
+    """
+    c, s = np.cos(angle), np.sin(angle)
+    if axis.lower() == 'x':
+        return np.matrix([[1, 0, 0], [0, c, s], [0, -s, c]], dtype=np.float32)
+    elif axis.lower() == 'y':
+        return np.matrix([[c, 0, s], [0, 1, 0], [-s, 0, c]], dtype=np.float32)
+    else:
+        return np.matrix([[c, -s, 0], [s, c, 0], [0, 0, 1]], dtype=np.float32)
+
+
+def rotate(m, x, y, z):
+    """
+    Perform a rotation given a rotation matrix and x, y, z coordinates (which can be arrays)
+    :param m:
+    :param x:
+    :param y:
+    :param z:
+    :return: a tuple of x, y, z coordinated after rotation
+    """
+    return m[0, 0] * x + m[0, 1] * y + m[0, 2] * z, \
+           m[1, 0] * x + m[1, 1] * y + m[1, 2] * z, \
+           m[2, 0] * x + m[2, 1] * y + m[2, 2] * z
+
 
 class Bragg2DPtychoData(object):
     """Class for two-dimensional ptychographic data: observed diffraction and probe positions.
     This may include only part of the data from a larger dataset.
     """
 
-    def __init__(self, iobs=None, positions=None, mask=None, wavelength=None, detector=None):
+    def __init__(self, iobs=None, positions=None, mask=None, wavelength=None, detector=None, scattering_vector=None):
         """
+        Init function. The orthonormal coordinate system to be used for the sample position follows the
+        NeXus/CXI/McStas convention:
+        - z parallel to the incident beam, downstream
+        - y: perpendicular to the direct beam and up
+        - x: perpendicular to the direct beam and towards the left when seen from the source.
 
-        :param iobs: 3d array with (nb_frame, ny,nx) shape observed intensity (assumed to follow Poisson statistics)
-        :param positions: (z, y, x) tuple or 2d array with ptycho probe positions in meters.
-                          The coordinates must follow the NeXus/CXI convention, and must already be in a
-                          laboratory frame taking into account sample rotation.
+        :param iobs: 3d array with (nb_frame, ny,nx) shape observed intensity (assumed to follow Poisson statistics).
+                     Data is assumed to be centered on the detector, and will be fft-shifted to be centered in (0,0).
+        :param positions: (z, y, x) tuple or 2d array with ptycho probe positions in meters. The coordinates
+                          must be in the laboratory reference frame given above.
         :param mask: 2D mask (>0 means masked pixel) for the observed data. Can be None.
         :param wavelength: wavelength of the experiment, in meters.
-        :param detector= {'geometry':'psic', 'delta':0, 'nu':0, 'pixel_size':55e-6, 'distance':1}:
-               parameters for the detector as a dictionary.
+        :param detector= {rotation_axes:(('x', 0), ('y', pi/4)), 'pixel_size':55e-6, 'distance':1}:
+               parameters for the detector as a dictionary. The rotation axes giving the detector orientation
+               will be applied in order, i.e. to find the detector direction, a vector with x=y=0, z=1 is rotated
+               by the axes in the given order.
+        :param scattering_vector=(hz, hy, hx): the coordinates of the central scattering vector for all frames, taking
+               into account any tilt for multi-angle Bragg projection ptychography. The scattering vector coordinates
+               units should be inverse meters (without 2pi multiplication), i.e. the norm should be 2*sin(theta)/lambda,
+               where theta is the Bragg angle. Note that only the difference between the scattering vector and the
+               average (intensity-weighted) scattering vector is used for multi-orientation back-projection.
+               [Default: None, all frames are taken with the same orientation]
         """
+        # TODO: alternatively, instead of supplying one detector geometry and scattering vectors which can vary from
+        # frame to frame, rather allow some geometry parameters to be vectors to be more general ? This
+        # would require including detector crop coordinates and rotation axes...
+
+        if scattering_vector is not None:
+            hz, hy, hx = scattering_vector
+            hx0 = hx.mean()
+            hy0 = hy.mean()
+            hz0 = hz.mean()
+            if iobs is not None:
+                iobs_sum = iobs.sum(axis=(1, 2))
+                s = iobs_sum.sum()
+                if s > 1000:  # Avoid the special case with null data for simulation
+                    hx0 = (hx * iobs_sum).sum() / s
+                    hy0 = (hy * iobs_sum).sum() / s
+                    hz0 = (hz * iobs_sum).sum() / s
+            self.s0 = (hx0, hy0, hz0)  # Average scattering vector
+            self.ds = (hx - hx0, hy - hy0, hz - hz0)  # Difference with average scattering vector
+        else:
+            self.s0 = (0, 0, 0)
+            tmp = np.zeros(len(iobs), dtype=np.float32)
+            self.ds = tmp.copy(), tmp.copy(), tmp.copy()
+
         if iobs is not None:
             self.iobs = np.fft.fftshift(iobs, axes=(-2, -1)).astype(np.float32)
             # Total nb of photons is used for regularization
             self.iobs_sum = self.iobs.sum()
             if mask is not None:
-                self.mask = np.fft.fftshift(mask.astype(np.int8))
-                self.iobs[mask > 0] = -100
+                mask3D = np.repeat(mask[np.newaxis, :, :], iobs.shape[0],
+                                   axis=0)  # tile mask to be a 3D array with (nb_frame, ny, nx) shape
+                self.mask = np.fft.fftshift(mask3D.astype(np.int8), axes=(-2, -1))
+                self.iobs[self.mask > 0] = -100
             else:
-                    self.mask = None
+                self.mask = None
         else:
             self.iobs = None
 
@@ -51,9 +120,13 @@ class Bragg2DPtychoData(object):
         self.posy -= self.posy.mean()
         self.posz -= self.posz.mean()
         self.detector = detector
-        lambdaz =  wavelength * detector['distance']
-        self.pixel_size_x = lambdaz / (detector['pixel_size'] * iobs.shape[-1])
-        self.pixel_size_y = lambdaz / (detector['pixel_size'] * iobs.shape[-2])
+        self.m = np.matrix([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float32)
+        for ax, angle in detector['rotation_axes']:
+            self.m *= rotation_matrix(ax, angle)
+        self.im = np.linalg.inv(self.m)
+
+        # Calculate ds coordinate in detector reference frame
+        self.ds1 = rotate(self.im, self.ds[0], self.ds[1], self.ds[2])
 
 
 class Bragg2DPtycho(object):
@@ -61,7 +134,7 @@ class Bragg2DPtycho(object):
     This may include only part of the data from a larger dataset
     """
 
-    def __init__(self, probe=None, data=None, support=None):
+    def __init__(self, probe=None, data=None, support=None, background=None):
         """
         Constructor.
         :param probe: the starting estimate of the probe, as a pynx wavefront object - can be 3D if modes are used.
@@ -73,14 +146,19 @@ class Bragg2DPtycho(object):
         self._obj = None
         self.support = support
         self.data = data
-        self._background = None
+        self._background = background
 
         # Matrix transformation from array indices of the array obtained by inverse Fourier Transform
-        # to xyz in the laboratory frame
+        # to xyz in the laboratory frame. Different from self.data.m, which is a pure rotation matrix
         self.m = None
         # Inverse of self.m
         self.im = None
 
+        # Voxel size in object and probe space
+        self.px1 = None
+        self.py1 = None
+        self.pz1 = None
+
         # Stored variables
         # self.scan_area_obj = None
         # self.scan_area_probe = None
@@ -100,6 +178,42 @@ class Bragg2DPtycho(object):
         self._cu_timestamp_counter = 0
         self.prepare()
 
+    def from_pu(self):
+        """
+        Get all relevant arrays from processing unit, if necessary
+        :return: Nothing
+        """
+        if self._cpu_timestamp_counter < self._timestamp_counter:
+            if self._timestamp_counter == self._cl_timestamp_counter:
+                if has_attr_not_none(self, '_cl_obj'):
+                    self._obj = self._cl_obj.get()
+                if has_attr_not_none(self, '_cl_probe'):
+                    self._probe = self._cl_probe.get()
+                if self._background is not None:
+                    if has_attr_not_none(self, '_cl_background'):
+                        self._background = self._cl_background.get()
+            if self._timestamp_counter == self._cu_timestamp_counter:
+                if has_attr_not_none(self, '_cu_obj'):
+                    self._obj = self._cu_obj.get()
+                if has_attr_not_none(self, '_cu_probe'):
+                    self._probe = self._cu_probe.get()
+                if self._background is not None:
+                    if has_attr_not_none(self, '_cu_background'):
+                        self._background = self._cu_background.get()
+            self._cpu_timestamp_counter = self._timestamp_counter
+
+    def get_probe(self):
+        """
+        Get the probe data array. This will automatically get the latest data, either from GPU or from the host
+        memory, depending where the last changes were made.
+
+        :param shift: if True, the data array will be fft-shifted so that the center of the data is in the center
+                      of the array, rather than in the corner (the default).
+        :return: the 3D numpy data array (nb object modes, nyo, nxo)
+        """
+        self.from_pu()
+        return self._probe
+
     def get_obj(self):
         """
         Get the object data array. This will automatically get the latest data, either from GPU or from the host
@@ -109,15 +223,7 @@ class Bragg2DPtycho(object):
                       of the array, rather than in the corner (the default).
         :return: the 3D numpy data array (nb object modes, nyo, nxo)
         """
-        if self._cpu_timestamp_counter < self._timestamp_counter:
-            if self._timestamp_counter == self._cl_timestamp_counter:
-                self._obj = self._cl_obj.get()
-                self._probe = self._cl_probe.get()
-            if self._timestamp_counter == self._cu_timestamp_counter:
-                self._obj = self._cu_obj.get()
-                self._probe = self._cu_probe.get()
-            self._cpu_timestamp_counter = self._timestamp_counter
-
+        self.from_pu()
         return self._obj
 
     def set_obj(self, obj):
@@ -136,49 +242,97 @@ class Bragg2DPtycho(object):
         self._timestamp_counter += 1
         self._cpu_timestamp_counter = self._timestamp_counter
 
-    def set_support(self, sup):
+        # Test if object and support have the same dimensions
+        if self.support is not None:
+            assert self.support.shape == self._obj.shape[-3:]
+
+    def set_support(self, sup, shrink_object_around_support=True):
         """
         Set the support data array. This should be a 3D array of the correct shape.
 
         :param sup: the support array. 0 outside support, 1 inside
+        :param shrink_object_around_support: if True, will shrink the object and support array around the support
+                                             volume. Note that the xyz coordinate may be shifted as a result, if
+                                             the support was not centered.
         :return: nothing
         """
-        self.support = sup.astype(np.int8)
-        self.support_sum = self.support.sum()
+        if sup is not None:
+            self.support = sup.astype(np.int8)
+            self.support_sum = self.support.sum()
+            self._timestamp_counter += 1
+            self._cpu_timestamp_counter = self._timestamp_counter
+            if shrink_object_around_support:
+                self.shrink_object_around_support()
+
+    def shrink_object_around_support(self):
+        """
+        Shrink the object around the tight support, to minimise the 3D object & support volume.
+        :return:
+        """
+        nzo, nyo, nxo = self._obj.shape[-3:]
+        z0, z1 = np.nonzero(self.support.sum(axis=(1, 2)))[0].take([0, -1])
+        y0, y1 = np.nonzero(self.support.sum(axis=(0, 2)))[0].take([0, -1])
+        x0, x1 = np.nonzero(self.support.sum(axis=(0, 1)))[0].take([0, -1])
+        self.set_support(self.support[z0:z1 + 1, y0:y1 + 1, x0:x1 + 1], shrink_object_around_support=False)
+        if self._obj is not None:
+            self.set_obj(self._obj[:, z0:z1 + 1, y0:y1 + 1, x0:x1 + 1])
+
+    def set_background(self, background):
+        """
+        Set the incoherent background data array.
+
+        :param background: the background (float32 numpy array)
+        :return: nothing
+        """
+        self._background = background.astype(np.float32)
         self._timestamp_counter += 1
         self._cpu_timestamp_counter = self._timestamp_counter
 
+    def get_background(self, shift=False):
+        """
+        Get the background data array. This will automatically get the latest data, either from GPU or from the host
+        memory, depending where the last changes were made.
+
+        :param shift: if True, the data array will be fft-shifted so that the center of the data is in the center
+                      of the array, rather than in the corner (the default).
+        :return: the 2D numpy data array
+        """
+        self.from_pu()
+        return self._background
+
     def prepare(self):
         """
         Calculate projection parameters
         :return:
         """
-        self.calc_orthonormalization_matrix()
+        self.calc_orthonormalisation_matrix()
         self.init_probe()
         self.init_obj()
+        self.set_support(self.support)
 
-    def calc_orthonormalization_matrix(self):
+    def calc_orthonormalisation_matrix(self):
         """
-        Calculate the orthonormalization matrix to convert probe/object array coordinates to/from orthonormal ones.
+        Calculate the orthonormalisation matrix to convert probe/object array coordinates (pixel coordinates in the
+        detector reference frame) to/from orthonormal ones in the laboratory reference frame.
+        This also initialises the voxel sizes in object and probe space
         :return:
         """
         npos, ny, nx = self.data.iobs.shape
-        if self.data.detector['geometry'] == 'psic':
-            cd, sd = np.cos(self.data.detector['delta']), np.sin(self.data.detector['delta'])
-            cn, sn = np.cos(self.data.detector['nu']), np.sin(self.data.detector['nu'])
-            # Pixel size. Assumes square frames from detector.
-            d = self.data.detector['distance']
-            lambdaz = self.data.wavelength * d
-            p = self.data.detector['pixel_size']
-            py, px = lambdaz / (p * ny), lambdaz / (p * nx)
-            pz = min(py / max(sd, 1e-10), px / max(sn, 1e-10))
-            self.m = np.array([[-cn * px, -sn * sd * py, -sn * cd * pz],
-                               [0, -cd * py, sd * pz],
-                               [-sn * px, sd * cn * py, cd * cn * pz]], dtype=np.float32)
-
-            self.im = np.linalg.inv(self.m)
-        else:
-            raise Exception("BraggPtycho: only recognized geometry is psic")
+        # Pixel size. Assumes square frames from detector.
+        d = self.data.detector['distance']
+        lambdaz = self.data.wavelength * d
+        p = self.data.detector['pixel_size']
+        py, px = lambdaz / (p * ny), lambdaz / (p * nx)
+        # TODO: Find method to estimate optimal pixel size along z. Would depend on probe shape and multiple angles..
+        pz = max(px, py)
+        self.m = self.data.m.copy()
+        self.m[:, 0] *= -px  # - sign because image pixels origin is 'top, left' seen from sample
+        self.m[:, 1] *= -py  # - sign because image pixels origin is 'top, left' seen from sample
+        self.m[:, 2] *= pz
+        self.im = np.linalg.inv(self.m)
+        self.px1 = px
+        self.py1 = py
+        self.pz1 = pz
 
     def init_obj(self):
         """
@@ -200,31 +354,102 @@ class Bragg2DPtycho(object):
         pixel_size_probe = self._probe2d.pixel_size
         # Extent of the probe along x and y in the laboratory reference frame
         nyp, nxp = self._probe2d.get().shape[-2:]
-        dy, dx = nyp // 2 * pixel_size_probe, nxp // 2 * pixel_size_probe
+        dyp, dxp = nyp // 2 * pixel_size_probe, nxp // 2 * pixel_size_probe
         # How far must we extend the back-propagated 2D wavefront along the sample-detector axis to go beyond
         # the probe path ? Use the corners of the projected wavefront to test
-        x, y, z = self.xyz_from_obj(np.array([0, 0, nx, nx]), np.array([0, ny, 0, ny]), 0)
-        if self.data.detector['geometry'] == 'psic':
-            cd, sd = np.cos(self.data.detector['delta']), np.sin(self.data.detector['delta'])
-            cn, sn = np.cos(self.data.detector['nu']), np.sin(self.data.detector['nu'])
-            # Pixel size along z
-            a = self.data.wavelength * self.data.detector['distance'] / self.data.detector['pixel_size']
-            p = self.data.detector['pixel_size']
-            py, px = a / ny, a / nx
-            pz = min(py / max(sd, 1e-10), px / max(sn, 1e-10))
-            # When are all corners outside propagate probe along x and y ? (z and -z should be symmetrical)
-            izmax_x = abs((dx - nx * px) / (pz * max(abs(sn * cd), 1e-10))).max()
-            izmax_y = abs((dy - ny * py) / (pz * max(abs(sd), 1e-10))).max()
-            nz = min(izmax_x, izmax_y) * 2
+        nx2 = nx / 2
+        ny2 = ny / 2
+        x, y, z = self.xyz_from_obj(np.array([-nx2, -nx2, nx2, nx2]), np.array([-ny2, ny2, -ny2, ny2]), 0)
+        d20 = self.data.m[0, 2]
+        if abs(d20) < 1e-10:
+            d20 = 1e-10
+        d21 = self.data.m[1, 2]
+        if abs(d21) < 1e-10:
+            d21 = 1e-10
+        # print("d20", d20, "d21", d21)
+
+        if d20 > 0:
+            # corner with lowest x, intersection with upper x probe border: x + alpha*d20 > dxp
+            i = x.argmin()
+            xc, yc, zc = x[i], y[i], z[i]
+            alpha = (dxp - xc) / d20
+            # print("alpha_max_x: ", alpha, i)
+            izmax_x = self.xyz_to_obj(xc + alpha * self.data.m[0, 2], yc + alpha * self.data.m[1, 2],
+                                      zc + alpha * self.data.m[2, 2])[2]
+
+            # corner with highest x, intersection with lower x probe border: x + alpha*d20 < -dxp
+            i = x.argmax()
+            xc, yc, zc = x[i], y[i], z[i]
+            alpha = (-dxp - xc) / d20
+            # print("alpha_min_x: ", alpha, i)
+            izmin_x = self.xyz_to_obj(xc + alpha * self.data.m[0, 2], yc + alpha * self.data.m[1, 2],
+                                      zc + alpha * self.data.m[2, 2])[2]
         else:
-            raise Exception("BraggPtycho: only recognized geometry is psic")
+            # corner with highest x, intersection with lower x probe border: x + alpha*d20 > -dxp
+            i = x.argmax()
+            xc, yc, zc = x[i], y[i], z[i]
+            alpha = (-dxp - xc) / d20
+            # print("alpha_max_x: ", alpha, i)
+            izmax_x = self.xyz_to_obj(xc + alpha * self.data.m[0, 2], yc + alpha * self.data.m[1, 2],
+                                      zc + alpha * self.data.m[2, 2])[2]
+
+            # corner with lowest x, intersection with higher x probe border: x + alpha*d20 < dxp
+            i = x.argmin()
+            xc, yc, zc = x[i], y[i], z[i]
+            alpha = (dxp - xc) / d20
+            # print("alpha_min_x: ", alpha, i)
+            izmin_x = self.xyz_to_obj(xc + alpha * self.data.m[0, 2], yc + alpha * self.data.m[1, 2],
+                                      zc + alpha * self.data.m[2, 2])[2]
+
+        # Same along y
+        if d21 > 0:
+            # corner with lowest y, intersection with upper y probe border: y + alpha*d21 > dyp
+            i = y.argmin()
+            xc, yc, zc = x[i], y[i], z[i]
+            alpha = (dyp - yc) / d21
+            # print("alpha_max_y: ", alpha, i)
+            izmax_y = self.xyz_to_obj(xc + alpha * self.data.m[0, 2], yc + alpha * self.data.m[1, 2],
+                                      zc + alpha * self.data.m[2, 2])[2]
+
+            # corner with highest y, intersection with lower y probe border: y + alpha*d21 < -dyp
+            i = y.argmax()
+            xc, yc, zc = x[i], y[i], z[i]
+            alpha = (-dyp - yc) / d21
+            # print("alpha_max_y: ", alpha, i)
+            izmin_y = self.xyz_to_obj(xc + alpha * self.data.m[0, 2], yc + alpha * self.data.m[1, 2],
+                                      zc + alpha * self.data.m[2, 2])[2]
+        else:
+            # corner with highest y, intersection with lower y probe border: y + alpha*d21 > -dyp
+            i = y.argmax()
+            xc, yc, zc = x[i], y[i], z[i]
+            alpha = (-dyp - yc) / d21
+            # print("alpha_max_y: ", alpha, i)
+            izmax_y = self.xyz_to_obj(xc + alpha * self.data.m[0, 2], yc + alpha * self.data.m[1, 2],
+                                      zc + alpha * self.data.m[2, 2])[2]
+
+            # corner with lowest y, intersection with higher y probe border: y + alpha*d21 < dyp
+            i = y.argmin()
+            xc, yc, zc = x[i], y[i], z[i]
+            alpha = (dyp - yc) / d21
+            # print("alpha_min_y: ", alpha, i)
+            izmin_y = self.xyz_to_obj(xc + alpha * self.data.m[0, 2], yc + alpha * self.data.m[1, 2],
+                                      zc + alpha * self.data.m[2, 2])[2]
+        # print("izmin_x: ", izmin_x)
+        # print("izmax_x: ", izmax_x)
+        # print("izmin_y: ", izmin_y)
+        # print("izmax_y: ", izmax_y)
+        # Final interval is smallest between x and y
+        izmax = min(izmax_x, izmax_y)
+        izmin = max(izmin_x, izmin_y)
+        nz = izmax - izmin
         print("Calculated probe shape: ", nz, ny, nx)
         return nz, ny, nx
 
     def calc_obj_shape(self, margin=8, multiple=2):
         """
         Calculate the 3D object shape, given the detector, probe and scan characteristics.
-        This must be called after the 3D probe has been initialized:
+        This must be called after the 3D probe has been initialized. Note that the final object shape will be shrunk
+        around the support once it is given.
         :param margin: margin to extend the object area, in case the positions will change (optimization)
         :param multiple: the shape must be a multiple of that number. >=2
         :return: the 3D object shape (nzo, nyo, nxo)
