@@ -10,7 +10,6 @@ import types
 import timeit
 import time
 from sys import stdout
-import os
 import warnings
 import numpy as np
 from matplotlib.backends.backend_agg import FigureCanvasAgg
@@ -124,10 +123,10 @@ class CPUOperatorPtycho(OperatorPtycho):
                 need_calc_q = False
 
         if need_calc_q:
-            x = np.fft.fftshift(np.arange(-nx // 2, nx // 2)).astype(np.complex64)
-            y = np.fft.fftshift(np.arange(-ny // 2, ny // 2)).astype(np.complex64)[:, np.newaxis]
+            x = np.fft.fftfreq(nx) * nx
+            y = np.fft.fftfreq(ny)[:, np.newaxis] * ny
 
-            q = np.exp(1j * abs(factor) * (x * x + y * y))
+            q = np.fft.fftshift(np.exp(1j * abs(factor) * (x * x + y * y)))
             p._cpu_tmp_quad_phase = (q.astype(np.complex64), abs(factor))
 
     def timestamp_increment(self, p):
@@ -279,10 +278,15 @@ class ShowObjProbe(CPUOperatorPtycho):
         self.title = title
 
     def op(self, p):
-        show_obj_probe(p.get_obj(), p.get_probe(), stit=self.title, fig_num=self.fig_num,
-                       pixel_size_object=p.pixel_size_object,
-                       scan_area_obj=p.scan_area_obj, scan_area_probe=p.scan_area_probe,
-                       scan_pos=p.scan_area_points)
+        if p.data.near_field:
+            show_obj_probe(p.get_obj(), p.get_probe(), stit=self.title, fig_num=self.fig_num,
+                           pixel_size_object=p.pixel_size_object, scan_area_obj=None, scan_area_probe=None,
+                           scan_pos=p.scan_area_points, minimize_obj_phase=False)
+        else:
+            show_obj_probe(p.get_obj(), p.get_probe(), stit=self.title, fig_num=self.fig_num,
+                           pixel_size_object=p.pixel_size_object,
+                           scan_area_obj=p.scan_area_obj, scan_area_probe=p.scan_area_probe,
+                           scan_pos=p.scan_area_points)
         return p
 
     def timestamp_increment(self, p):
@@ -368,12 +372,12 @@ class ObjProbe2Psi(CPUOperatorPtycho):
         obs = p.data.iobs
         n_stack = len(obs)
         p._psi = np.empty((nb_obj, nb_probe, n_stack, ny, nx), dtype=np.complex64)
-        py, px = p.data.pixel_size_object()
+        px, py = p.data.pixel_size_object()
         for iobj in range(len(p._obj)):
             for iprobe in range(len(p._probe)):
                 for i in range(n_stack):
                     dy, dx = p.data.posy[i] / py, p.data.posx[i] / px
-                    cy, cx = get_view_coord((nyo, nxo), (ny, nx), dy, dx)
+                    cx, cy = get_view_coord((nyo, nxo), (ny, nx), dx, dy)
                     op = p._probe[iprobe] * p._obj[iobj, cy:cy + ny, cx:cx + nx]
                     if p.data.near_field is False:
                         op *= p._cpu_tmp_quad_phase[0]
@@ -569,20 +573,14 @@ class LLK(CPUOperatorPtycho):
     """
 
     def op(self, p: Ptycho):
-        iobs = p.data.iobs
-        icalc = (np.abs(p._psi) ** 2).sum(axis=(0, 1))
-        ny, nx = iobs.shape[-2:]
-
-        icalc = icalc * (iobs < 0) + iobs * (iobs >= 0)
-
-        iobs = iobs.flatten()
-        icalc = icalc.flatten()
+        iobs = p.data.iobs.flatten()
+        icalc = (np.abs(p._psi) ** 2).sum(axis=(0, 1)).flatten()
 
         # Poisson
-        llk = np.empty(iobs.shape, dtype=np.float32)
+        llk = np.zeros(iobs.shape, dtype=np.float32)
         idx = np.where(iobs > 0)
         llk[idx] = np.take(icalc - iobs + iobs * np.log(iobs / icalc), idx)
-        idx = np.where(iobs <= 0)
+        idx = np.where(iobs == 0)
         llk[idx] = np.take(icalc, idx)
         p.llk_poisson = llk.sum()
 
@@ -622,7 +620,7 @@ class Psi2ObjProbe(CPUOperatorPtycho):
         q = p._cpu_tmp_quad_phase[0]
         obs = p.data.iobs
         n_stack = len(obs)
-        py, px = p.data.pixel_size_object()
+        px, py = p.data.pixel_size_object()
         if self.update_object:
             obj_new = np.zeros_like(p._obj)
             obj_norm = np.zeros_like(p._obj[0], dtype=np.float32)
@@ -631,7 +629,7 @@ class Psi2ObjProbe(CPUOperatorPtycho):
                     pr = p._probe[iprobe]
                     for i in range(n_stack):
                         dy, dx = p.data.posy[i] / py, p.data.posx[i] / px
-                        cy, cx = get_view_coord((nyo, nxo), (ny, nx), dy, dx)
+                        cx, cy = get_view_coord((nyo, nxo), (ny, nx), dx, dy)
                         obj_new[iobj, cy:cy + ny, cx:cx + nx] += np.fft.fftshift(
                             p._psi[iobj, iprobe, i] / q) * pr.conj()
                         obj_norm[cy:cy + ny, cx:cx + nx] += abs(pr) ** 2
@@ -647,7 +645,7 @@ class Psi2ObjProbe(CPUOperatorPtycho):
                 for iprobe in range(len(p._probe)):
                     for i in range(n_stack):
                         dy, dx = p.data.posy[i] / py, p.data.posx[i] / px
-                        cy, cx = get_view_coord((nyo, nxo), (ny, nx), dy, dx)
+                        cx, cy = get_view_coord((nyo, nxo), (ny, nx), dx, dy)
                         ob = p._obj[iobj, cy:cy + ny, cx:cx + nx]
                         p._probe[iprobe] += np.fft.fftshift(p._psi[iobj, iprobe, i] / q) * ob.conj()
                         probe_norm += abs(ob) ** 2
@@ -665,12 +663,14 @@ class AP(CPUOperatorPtycho):
     - update background optionally
     """
 
-    def __init__(self, update_object=True, update_probe=False, update_background=False, nb_cycle=1, calc_llk=False,
-                 show_obj_probe=False, fig_num=-1):
+    def __init__(self, update_object=True, update_probe=False, update_background=False, floating_intensity=False,
+                 nb_cycle=1, calc_llk=False, show_obj_probe=False, fig_num=-1):
         """
 
         :param update_object: update object ?
         :param update_probe: update probe ?
+        :param update_background: update background ? [TODO CPU operators]
+        :param floating_intensity: optimise floating intensity scale factor [TODO CPU operators]
         :param nb_cycle: number of cycles to perform. Equivalent to AP(...)**nb_cycle
         :param calc_llk: if True, calculate llk while in Fourier space. If a positive integer is given, llk will be
                          calculated every calc_llk cycle
@@ -682,6 +682,7 @@ class AP(CPUOperatorPtycho):
         self.update_object = update_object
         self.update_probe = update_probe
         self.update_background = update_background
+        self.floating_intensity = floating_intensity  # TODO
         self.nb_cycle = nb_cycle
         self.calc_llk = calc_llk
         self.show_obj_probe = show_obj_probe
@@ -694,7 +695,8 @@ class AP(CPUOperatorPtycho):
         :return: a new DM operator with the number of cycles multiplied by n
         """
         assert isinstance(n, int) or isinstance(n, np.integer)
-        return AP(update_object=self.update_object, update_probe=self.update_probe, nb_cycle=self.nb_cycle * n,
+        return AP(update_object=self.update_object, update_probe=self.update_probe,
+                  floating_intensity=self.floating_intensity, nb_cycle=self.nb_cycle * n,
                   calc_llk=self.calc_llk, show_obj_probe=self.show_obj_probe, fig_num=self.fig_num)
 
     def op(self, p: Ptycho):
@@ -717,10 +719,14 @@ class AP(CPUOperatorPtycho):
                 ic_dt = ic + 1
                 t0 = timeit.default_timer()
 
-                s = algo_string('AP', p, self.update_object, self.update_probe, self.update_background)
-                print("%s #%3d LLK= %8.2f(p) %8.2f(g) %8.2f(e), nb photons=%e, dt/cycle=%5.3fs" \
-                      % (s, ic, p.llk_poisson / p.nb_obs, p.llk_gaussian / p.nb_obs, p.llk_euclidian / p.nb_obs,
-                         p.nb_photons_calc, dt))
+                p.update_history(mode='llk', update_obj=self.update_object, update_probe=self.update_probe,
+                                 update_background=self.update_background, update_pos=False, dt=dt, algorithm='AP',
+                                 verbose=True)
+            else:
+                p.history.insert(p.cycle, update_obj=self.update_object, update_probe=self.update_probe,
+                                 update_background=self.update_background, update_pos=False, algorithm='AP',
+                                 verbose=False)
+            p.cycle += 1
 
             if self.show_obj_probe:
                 if ic % self.show_obj_probe == 0 or ic == self.nb_cycle - 1:
@@ -738,9 +744,10 @@ class DM(CPUOperatorPtycho):
     """
 
     def __init__(self, update_object=True, update_probe=True, nb_cycle=1, calc_llk=False, show_obj_probe=False,
-                 fig_num=-1):
+                 fig_num=-1, obj_smooth_sigma=0, obj_inertia=0.01, probe_smooth_sigma=0, probe_inertia=0.001,
+                 center_probe_n=0, center_probe_max_shift=5, loop_obj_probe=1):
         """
-
+        TODO: Not all parameters are implemented in the CPU DM version.
         :param update_object: update object ?
         :param update_probe: update probe ?
         :param nb_cycle: number of cycles to perform. Equivalent to DM(...)**nb_cycle
@@ -806,18 +813,22 @@ class DM(CPUOperatorPtycho):
                 ic_dt = ic + 1
                 t0 = timeit.default_timer()
 
-                s = algo_string('DM', p, self.update_object, self.update_probe)
-                print("%s #%3d LLK= %8.2f(p) %8.2f(g) %8.2f(e), nb photons=%e, dt/cycle=%5.3fs" \
-                      % (s, ic, p.llk_poisson / p.nb_obs, p.llk_gaussian / p.nb_obs, p.llk_euclidian / p.nb_obs,
-                         p.nb_photons_calc, dt))
+                p.update_history(mode='llk', update_obj=self.update_object, update_probe=self.update_probe,
+                                 update_background=False, update_pos=False, dt=dt, algorithm='DM',
+                                 verbose=True)
                 # Restore correct Psi
                 p._psi = psi1
+            else:
+                p.history.insert(p.cycle, update_obj=self.update_object, update_probe=self.update_probe,
+                                 update_background=False, update_pos=False, algorithm='DM',
+                                 verbose=False)
 
             if self.show_obj_probe:
                 if ic % self.show_obj_probe == 0 or ic == self.nb_cycle - 1:
                     s = algo_string('DM', p, self.update_object, self.update_probe)
                     tit = "%s #%3d, LLKn(p)=%8.3f" % (s, ic, p.llk_poisson / p.nb_obs)
                     p = ShowObjProbe(fig_num=self.fig_num, title=tit) * p
+            p.cycle += 1
 
         return p
 
@@ -912,10 +923,10 @@ class ML(CPUOperatorPtycho):
             # 1) Compute the gradients (actually the gradients conjugates)
             p = Propagate(forward=True) * ObjProbe2Psi() * p
             calc = (abs(p._psi) ** 2).sum(axis=(0, 1))
-            p._psi *= (1 - obs / calc)  # with broadcasting
+            p._psi *= (1 - obs / calc) * (obs >= 0)  # with broadcasting
             p = Propagate(forward=False) * p
 
-            py, px = p.data.pixel_size_object()
+            px, py = p.data.pixel_size_object()
             if self.update_object:
                 obj_grad[:] = 0
                 for iobj in range(nb_obj):
@@ -923,7 +934,7 @@ class ML(CPUOperatorPtycho):
                         pr = p._probe[iprobe]
                         for i in range(n_stack):
                             dy, dx = p.data.posy[i] / py, p.data.posx[i] / px
-                            cy, cx = get_view_coord((nyo, nxo), (ny, nx), dy, dx)
+                            cx, cy = get_view_coord((nyo, nxo), (ny, nx), dx, dy)
                             obj_grad[iobj, cy:cy + ny, cx:cx + nx] += np.fft.fftshift(
                                 p._psi[iobj, iprobe, i] / p._cpu_tmp_quad_phase[0]) * pr.conj()
 
@@ -933,7 +944,7 @@ class ML(CPUOperatorPtycho):
                     for iprobe in range(len(p._probe)):
                         for i in range(n_stack):
                             dy, dx = p.data.posy[i] / py, p.data.posx[i] / px
-                            cy, cx = get_view_coord((nyo, nxo), (ny, nx), dy, dx)
+                            cx, cy = get_view_coord((nyo, nxo), (ny, nx), dx, dy)
                             ob = p._obj[iobj, cy:cy + ny, cx:cx + nx]
                             probe_grad[iprobe] += np.fft.fftshift(
                                 p._psi[iobj, iprobe, i] / p._cpu_tmp_quad_phase[0]) * ob.conj()
@@ -1055,16 +1066,20 @@ class ML(CPUOperatorPtycho):
                 ic_dt = ic + 1
                 t0 = timeit.default_timer()
 
-                s = algo_string('ML', p, self.update_object, self.update_probe, self.update_background)
-                print("%s #%3d LLK= %8.2f(p) %8.2f(g) %8.2f(e), nb photons=%e, dt/cycle=%5.3fs" \
-                      % (s, ic, p.llk_poisson / p.nb_obs, p.llk_gaussian / p.nb_obs, p.llk_euclidian / p.nb_obs,
-                         p.nb_photons_calc, dt))
+                p.update_history(mode='llk', update_obj=self.update_object, update_probe=self.update_probe,
+                                 update_background=self.update_background, update_pos=False, dt=dt, algorithm='ML',
+                                 verbose=True)
+            else:
+                p.history.insert(p.cycle, update_obj=self.update_object, update_probe=self.update_probe,
+                                 update_background=self.update_background, update_pos=False, algorithm='ML',
+                                 verbose=False)
 
             if self.show_obj_probe:
                 if ic % self.show_obj_probe == 0 or ic == self.nb_cycle - 1:
                     s = algo_string('ML', p, self.update_object, self.update_probe, self.update_background)
                     tit = "%s #%3d, LLKn(p)=%8.3f" % (s, ic, p.llk_poisson / p.nb_obs)
                     p = ShowObjProbe(fig_num=self.fig_num, title=tit) * p
+            p.cycle += 1
 
         return p
 
@@ -1074,6 +1089,14 @@ class ScaleObjProbe(CPUOperatorPtycho):
     Scale object and probe so that they have the same magnitude.
     """
 
+    def __init__(self, verbose=False):
+        """
+
+        :param verbose: print deviation if verbose=True
+        """
+        super(ScaleObjProbe, self).__init__()
+        self.verbose = verbose
+
     def op(self, p):
         # TODO: scale taking into account the mask ?
         nb_photons_obs = p.data.iobs.sum()
@@ -1084,7 +1107,8 @@ class ScaleObjProbe(CPUOperatorPtycho):
         ps = np.abs(p._probe).sum()
         p._probe *= np.sqrt(os / ps * s)
         p._obj *= np.sqrt(ps / os * s)
-        # print("ScaleObjProbe", ps, os, np.sqrt(os / ps), np.sqrt(ps / os))
+        if self.verbose:
+            print("ScaleObjProbe:", ps, os, s, np.sqrt(os / ps * s), np.sqrt(ps / os * s))
         return p
 
 
@@ -1191,7 +1215,7 @@ class AnalyseProbe(CPUOperatorPtycho):
                                                                do_plot=False)
 
             if self.verbose:
-                print("Probe statistics at found focus position (z=%+6.1fum):" % (vdz[izmax] * 1e6))
+                print("\nProbe statistics at found focus position (z=%+6.1fum):" % (vdz[izmax] * 1e6))
                 analysis.probe_fwhm(p3d[izmax], p.data.pixel_size_object()[0])
             if self.save_prefix is not None or self.show_plot:
                 dy = (6 + 1) / 72 / fig.get_size_inches()[1]
