@@ -14,9 +14,10 @@ import sys
 import warnings
 import numpy as np
 from numpy import pi
+from scipy.ndimage import zoom
 from matplotlib import pyplot as plt
 from ..utils.plot_utils import showCplx
-from ..utils.pattern import get_img
+from ..utils.pattern import get_img, spiral_fermat, spiral_archimedes
 from .shape import get_view_coord
 from ..wavefront import Wavefront, PropagateNearField, PropagateFarField, FreePU
 
@@ -130,48 +131,9 @@ def tukeywin2D(window_size, alpha=0.5):
     return np.outer(w1, w2)
 
 
-def spiral_archimedes(a, n):
-    """" Creates np points spiral of step a, with a between successive points
-    on the spiral. Returns the x,y coordinates of the spiral points.
-
-    This is an Archimedes spiral. the equation is:
-      r=(a/2*pi)*theta
-      the stepsize (radial distance between successive passes) is a
-      the curved absciss is: s(theta)=(a/2*pi)*integral[t=0->theta](sqrt(1*t**2))dt
-    """
-    vr, vt = [0], [0]
-    t = np.pi
-    while len(vr) < n:
-        vt.append(t)
-        vr.append(a * t / (2 * np.pi))
-        t += 2 * np.pi / np.sqrt(1 + t ** 2)
-    vt, vr = np.array(vt), np.array(vr)
-    return vr * np.cos(vt), vr * np.sin(vt)
-
-
-def spiral_fermat(dmax, n):
-    """"
-    Creates a Fermat spiral with n points distributed in a circular area with
-    diameter<= dmax. Returns the x,y coordinates of the spiral points. The average
-    distance between points can be roughly estimated as 0.5*dmax/(sqrt(n/pi))
-
-    http://en.wikipedia.org/wiki/Fermat%27s_spiral
-    """
-    c = 0.5 * dmax / np.sqrt(n)
-    vr, vt = [], []
-    t = .4
-    goldenAngle = np.pi * (3 - np.sqrt(5))
-    while t < n:
-        vr.append(c * np.sqrt(t))
-        vt.append(t * goldenAngle)
-        t += 1
-    vt, vr = np.array(vt), np.array(vr)
-    return vr * np.cos(vt), vr * np.sin(vt)
-
-
-def psi(obj, probe, dy, dx, border):
+def psi(obj, probe, dx, dy, border):
     # TODO: subpixel shift
-    cy, cx = get_view_coord(obj.shape, probe.shape, dy, dx)
+    cx, cy = get_view_coord(obj.shape, probe.shape, dx, dy)
     ny, nx = probe.shape
     return probe * obj[cy:cy + ny, cx:cx + nx]
 
@@ -274,7 +236,7 @@ class Simulation(object):
                                     scan_info['integer_values'] = True # integer values of the scan positions
                                     defaults set in update_default_scan
 
-        scan: specific value of scan can be passed (list/tuple of two arrays (posy, posx)). scan_info can be passed as empty dicitionary (default).
+        scan: specific value of scan can be passed (list/tuple of two arrays (posx, posy)). scan_info can be passed as empty dictionary (default).
 
         data_info: dictionary with data_info['pix_size_direct_nm'] = value # pixels size in the object space [nm]
                                     data_info['num_phot_max'] = value #number of photons in maximum pixel
@@ -377,7 +339,7 @@ class Simulation(object):
         self.obj.make_even()
         self.probe.make_even()
 
-        posy, posx = self.scan.values
+        posx, posy = self.scan.values
         posy_max, posx_max = np.ceil(abs(posy).max()), np.ceil(abs(posx).max())
         n = len(posx)  # number of scan positions
 
@@ -421,7 +383,7 @@ class Simulation(object):
                 if (n - index) % 10 == 0 or index == 0:
                     sys.stdout.write('%d ' % (n - index))
                     sys.stdout.flush()
-            self.psi.values[index] = psi(self.obj.values, self.probe.values, sy, sx, self.obj.info['border_pix'])
+            self.psi.values[index] = psi(self.obj.values, self.probe.values, sx, sy, self.obj.info['border_pix'])
             self.wd = Wavefront(d=np.fft.fftshift(self.psi.values[index]), wavelength=wavelength,
                                 pixel_size=pixel_size_object)
             if near_field:
@@ -466,10 +428,24 @@ class Simulation(object):
         # print("posx_max=%d, p_s[0] // 2=%d, border=%d" % (posx_max, p_s[0] // 2, border), ot_s)
         ot_s = max(ot_s[0], o_s[0]), max(ot_s[1], o_s[1])  # the case of very large object ot_s<o_s
         # print("posy_max=%d, p_s[1] // 2=%d, border=%d" % (posy_max, p_s[1] // 2, border), ot_s)
-        tmp = np.zeros(ot_s, np.dtype(self.obj.values.dtype))
-        start = ot_s[0] // 2 - o_s[0] // 2, ot_s[1] // 2 - o_s[1] // 2
-        tmp[start[0]:start[0] + o_s[0], start[1]: start[1] + o_s[1]] = self.obj.values
+
+        if False:
+            # Fill in center with image
+            tmp = np.zeros(ot_s, np.dtype(self.obj.values.dtype))
+            start = ot_s[0] // 2 - o_s[0] // 2, ot_s[1] // 2 - o_s[1] // 2
+            tmp[start[0]:start[0] + o_s[0], start[1]: start[1] + o_s[1]] = self.obj.values
+        else:
+            oa, op = np.abs(self.obj.values), np.angle(self.obj.values)
+            # Zoom image
+            tmpa = zoom(oa, (ot_s[0] / o_s[0], ot_s[1] / o_s[1]))
+            tmpp = zoom(op, (ot_s[0] / o_s[0], ot_s[1] / o_s[1]))
+            tmp = (tmpa * np.exp(1j * tmpp)).astype(self.obj.values.dtype)
+        old_info = self.obj.info
         self.obj = Im(tmp, info)
+        if len(old_info):
+            self.obj.info = old_info
+            if 'border_pix' not in self.obj.info:
+                self.obj.info['border_pix'] = 2
 
     def make_obj(self):
         self.update_default_obj()
@@ -484,7 +460,7 @@ class Simulation(object):
         elif ('ampl' in obj_type) and ('phase' in obj_type):
             im0 = self.DTYPE_REAL(get_img(0))
             im1 = self.DTYPE_REAL(get_img(1))
-            # Strethc the phase to interval (-phase_stretch/2, +phase_stretch/2)
+            # Stretch the phase to interval (-phase_stretch/2, +phase_stretch/2)
             phase0 = im1 - im1.min()
             ps = 2 * np.pi
             phase_stretch = ps * phase0 / self.DTYPE_REAL(phase0.max()) - ps / 2.
@@ -528,10 +504,15 @@ class Simulation(object):
                 phase_stretch = ps * phase0 / self.DTYPE_REAL(phase0.max()) - ps / 2.
                 obj = abs(obj) * np.exp(1j * phase_stretch)
 
+        if 'ampl_range' in info and not 'range' in info:
+            a0, a1 = info['ampl_range']
+            obj = (a0 + a1 * np.abs(obj) / np.abs(obj).max()) * np.exp(1j * np.angle(obj))
+
         if 'alpha_win' in info:
             s = obj.shape
+            mi = abs(obj).min()
             w = tukeywin2D(s, info['alpha_win'])
-            obj *= w  # tuckey window to smooth edges (alpha_win)
+            obj = mi + w * (obj - mi)  # tuckey window to smooth edges (alpha_win)
 
         self.obj.values = self.DTYPE_CPLX(obj)
 
@@ -568,7 +549,7 @@ class Simulation(object):
             pixel_size_focus = None
             if 'detector_distance' in self.data_info.keys() and 'detector_pixel_size' in self.data_info.keys():
                 pixel_size_focus = self.wavelength * self.data_info['detector_distance'] / (
-                    nx * self.data_info['detector_pixel_size'])
+                        nx * self.data_info['detector_pixel_size'])
             elif 'pix_size_direct_nm' in self.data_info.keys():
                 pixel_size_focus = self.data_info['pix_size_direct_nm'] * 1e-9
             assert pixel_size_focus is not None, "Could not determine pixel size at focus to calculate probe from focused aperture"
@@ -643,20 +624,22 @@ class Simulation(object):
 
         elif scan_type == 'spiral':
             posy, posx = spiral_archimedes(info['scan_step_pix'], info['n_scans'])
-        elif scan_type == 'spiralfermat':
-            posy, posx = spiral_archimedes(info['scan_step_pix'], info['n_scans'])
+        elif scan_type in ['spiralfermat', 'spiral_fermat']:
+            n = info['n_scans']
+            dmax = 2 * np.sqrt(n / np.pi) * info['scan_step_pix']
+            posy, posx = spiral_fermat(dmax, n)
         else:
             msg = "Unknown scan type:", self.scan.info['type']
             raise NameError(msg)
 
         if info['integer_values']:
             # print("Integer values of the scan!")
-            self.scan.values = (posy.round(), posx.round())
+            self.scan.values = (posx.round(), posy.round())
         else:
-            self.scan.values = (posy, posx)
+            self.scan.values = (posx, posy)
 
     def show_illumination_sum(self, log_values=False):
-        posy, posx = self.scan.values
+        posx, posy = self.scan.values
         illum = np.zeros_like(self.obj.values)
         nix, niy = illum.shape
         npx, npy = self.probe.values.shape
