@@ -153,7 +153,7 @@ class CLOperatorWavefront(OperatorWavefront):
         # Make sure data is already in OpenCL space, otherwise transfer it
         if w._timestamp_counter > w._cl_timestamp_counter:
             # print("Moving data to OpenCL space")
-            w._cl_d = pyopencl.array.to_device(self.processing_unit.cl_queue, w._d, async=False)
+            w._cl_d = pyopencl.array.to_device(self.processing_unit.cl_queue, w._d, async_=False)
             w._cl_timestamp_counter = w._timestamp_counter
         if has_attr_not_none(w, '_cl_d_view') is False:
             w._cl_d_view = {}
@@ -319,7 +319,7 @@ class ToPU(CLOperatorWavefront):
         :return: the updated Wavefront object
         """
         warnings.warn("Use of ToPU() and FromPU() operators is now deprecated. Use get() and set() to access data.")
-        # w._cl_d = pyopencl.array.to_device(self.processing_unit.cl_queue, w.d, async=False)
+        # w._cl_d = pyopencl.array.to_device(self.processing_unit.cl_queue, w.d, async_=False)
         return w
 
 
