@@ -294,6 +294,9 @@ class ImshowAbs(Imshow):
 
         :param fig_num: the matplotlib figure number. if None, a new figure will be created each time.
         :param i: if the wavefront is actually a stack of 2D wavefronts, display w.get()[i]
+        :param mode: either 'linear' or 'log', the scaling using for the colour amplitude. In both modes the display
+                     min and max are the 0.1 and 99.9 percentile of the abs or log(abs) of the entire array.
+        :param kwargs_complex2rgba: kwargs to be passed to complex2rgbalin or complex2rgbalog
         :param title: the title for the view. If None, a default title will be used.
         """
         super(ImshowAbs, self).__init__(fig_num=fig_num, i=i, title=title)
@@ -302,7 +305,16 @@ class ImshowAbs(Imshow):
 
     def op(self, w):
         d, x, y, unit_name = self.pre_imshow(w)
-        plt.imshow(np.abs(d), extent=(x.min(), x.max(), y.min(), y.max()), cmap=plt.cm.get_cmap('gray'))
+        if 'log' in self.mode:
+            a = np.log10(np.abs(d))
+            vmin, vmax = np.percentile(a, [0.1, 99.9])
+            plt.imshow(a, vmin=vmin, vmax=vmax, extent=(x.min(), x.max(), y.min(), y.max()),
+                       cmap=plt.cm.get_cmap('gray'))
+        else:
+            a = np.abs(d)
+            vmin, vmax = np.percentile(a, [0.1, 99.9])
+            plt.imshow(a, vmin=vmin, vmax=vmax, extent=(x.min(), x.max(), y.min(), y.max()),
+                       cmap=plt.cm.get_cmap('gray'))
         plt.colorbar()
         self.post_imshow(w, x, y, unit_name, title_append="[Amplitude]")
         return w
